
open Stree

type lexeme =
        | String of string
        | Operator of char
        | Indent of int
        | Pbreak

let specialcharp = function
        | ('(' | ')' | '%' | '&' | '$' | '#' | '+' | '-' | '/' | ':' | '<' | '='
          | '>' | '?' | '@' | '\\' | '~' | '^' | '|' | '*' | '.' | '"' | '['
          | ']' | '{' | '}' | '\'') -> true
        | _ -> false
let whitespacep = function
        | ' ' | '\t' | '\n' | '\r' -> true
        | _ -> false

let rec indent n = parser
        | [< '' '; rest >] -> indent (n+1) rest
        | [< ''\t'; rest >] -> indent (((n/8)+1)*8) rest
        | [< >] -> n
let poss_gap = parser
        | [< ''\n' >] -> true
        | [< >] -> false
let rec word s = parser
        | [< 'c when not (specialcharp c) && not (whitespacep c); rest >] ->
                        word (s ^ String.make 1 c) rest
        | [< >] -> s
let rec commentchars = parser
        | [< 'c when c <> '\n'; rest >] -> commentchars rest
        | [< >] -> ()
let rec tokenize = parser
        | [< ''\n'; g = poss_gap; n = indent 0; rest >] ->
                        if g then [< 'Pbreak; 'Indent n; tokenize rest >]
                        else [< 'Indent n; tokenize rest >]
        | [< '';'; c = commentchars; rest >] -> tokenize rest
        | [< 'c when whitespacep c; rest >] -> tokenize rest
        | [< 'c when specialcharp c; rest >] -> [< 'Operator c; tokenize rest >]
        | [< s = word ""; rest >] -> begin match s with 
                | "" -> [< >]
                | _ -> [< 'String s; tokenize rest >]
                end
let tokenize = parser
        | [< n = indent 0; t = tokenize >] -> [< 'Indent n; t >]

let normalise = function
        | [t] -> t
        | x -> Seq x
let rec horiz_seq = parser
        | [< 'String x when List.mem x ops; rest >] ->
                        List.assoc x op_tags :: horiz_seq rest
        | [< 'String s; rest >] -> Symbol s :: horiz_seq rest
        | [< 'Operator c; rest >] -> Symbol (String.make 1 c) :: horiz_seq rest
        | [< >] -> []
let rec sequence level = parser
        | [< 'Pbreak; rest >] -> Symbol "\n" :: sequence level rest
        | [< 'Indent n when n > level; s = sequence n; rest >] ->
                        Seq s :: sequence level rest
        | [< 'Indent n when n = level; s = horiz_seq; rest >] ->
                        normalise s :: sequence level rest
        | [< s = horiz_seq; rest >] -> begin match s with
                | [] -> []
                | _ -> normalise s :: sequence level rest
                end

let program = sequence 0

let from_string s = program (tokenize (Stream.of_string s))
let from_file f = program (tokenize (Stream.of_channel (open_in f)))

